/**
 * Class for dealing with static scoring.
 */

package org.ucam.srcf.assassins.strawberry;

import org.ucam.srcf.assassins.Death;
import org.ucam.srcf.assassins.AssassinViewer;
import org.ucam.srcf.assassins.Debug;
import org.ucam.srcf.assassins.exceptions.AssassinSQLException;

import org.netlib.lapack.DGETRF;
import org.netlib.lapack.DGETRS;
import org.netlib.util.intW;

import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;

class KirikaStatic implements StaticScoringAlgorithm {
   private double A[][];
   private double deaths[];
   private Map assIndex=new HashMap();
   private int next_index=0;
   private boolean dirtyMatrix=false;
   private static final int c = 6;

   public void init (int n) {
      A=new double[n][n];
      deaths=new double[n];
   }

   /** 
    * Wrapper around the munge and solve routine.
    * @param c Scoring constant.
    * @return Map from AssassinViewer to Double representing score.
    */
   public Map score() {
      double[] b, x;

      if(Debug.debug) {
         Debug.print(this, "Kill Matrix");
         printMatrix(next_index);
      }
      b=mungeMatrix(c);
      if(Debug.debug) {
         Debug.print(this, "Munged Matrix");
         printMatrix(next_index);
         Debug.print(this, "RHS Vector");
         printVector(b,next_index);
      }
      x=solveSystem(b);

      Map scores=new HashMap();
      Iterator i=assIndex.keySet().iterator();
      while(i.hasNext()) {
         AssassinViewer current=(AssassinViewer) i.next();
         int index=getIndex(current);
         scores.put(current, new Double(x[index]));
      }

      return scores;
   }

   private void printMatrix(int n) {
      for(int i=0; i<n; i++)
         printVector(A[i], n);
   }

   private void printVector(double[] x, int n) {
      String str=x[0]+"";
      for(int i=1; i<n; i++)
         str+=", "+x[i];
      Debug.print(this, str);
   }
   
   public void addDeath(Death d) throws AssassinSQLException {
      if(dirtyMatrix)
         throw new RuntimeException("Matrix already scored!");
      
      int kindex=getIndex(d.getKiller());
      int vindex=getIndex(d.getVictim());

      A[kindex][vindex]++;
   }

   private int getIndex(AssassinViewer a) {
      Integer index = (Integer) assIndex.get(a);
      if(index==null)
         assIndex.put(a, index = new Integer(next_index++));
      return index.intValue();
   }

   /**
    * This routine takes a multiple death count and levels it.
    * We use partial sums of the series (1/2^n)_{n\ge 0}.
    * @param n number of deaths
    */
   private double levelDeath(int n) {
      double sum=0;
      for(int i=0;i<n;i++)
         sum+=1.0/Math.pow(2,i);
      return sum;
   }

   /** 
    * Munge the matrix ready for solving.
    * @param c the scoring parameter - we transfer an "equivilant" 
    * of 1/6 points for a kill
    * @return vector to solve the system equal unto.
    */
   private double[] mungeMatrix(int c) {
      dirtyMatrix=true;
      /* Count number of deaths */
      for(int i=0; i<A.length; i++) {
         deaths[i]=0;
         for(int j=0;j<A.length;j++) {
            A[j][i]=levelDeath((int) A[j][i]);
            deaths[i]+=A[j][i];
         }
      }
      if(Debug.debug) {
         Debug.print(this, "Deaths vector");
         printVector(deaths, next_index);
         Debug.print(this, "Death Levelled Matrix");
         printMatrix(next_index);
      }

      /* (1-[(c-1)/c]^m)/deaths * kills */
      for(int i=0;i<A.length;i++) {
         if(deaths[i]==0)
            continue;
         for(int j=0;j<A.length;j++) {
            A[j][i]*=(1-Math.pow(((c-1.0)/c), deaths[i]));
            A[j][i]/=deaths[i];
         }
      }

      /* Put -1's down the daigonal */
      for(int i=0;i<A.length;i++) {
         A[i][i]=-1;
      }

      /* Find the vector we're solving for */
      double[] x=new double[A.length];
      for(int i=0;i<A.length;i++) {
         x[i]=-1;
         /*x[i]=-Math.pow((c-1.0)/c, deaths[i]);*/
      }

      return x;
   }

   /**
    * Solve the matrix using an efficient method.
    * Ax=b
    * Where A is assumed to be this.A
    * May solve b in place, but this is not garunteed.
    * @return x a soln to Ax=b
    */
   private double[] solveSystem(double[] b) {
      int n=A.length;
      intW info = new intW(0);
      int[] ipiv=new int[A.length];
      double[][] rhs=new double[n][1];

      DGETRF.DGETRF(n,n,A,ipiv,info);
      if(info.val!=0)
         throw new RuntimeException("Badness in DGETRF LAPACK call, info="+info);

      for(int i=0; i<n; i++)
         rhs[i][0]=b[i];
      DGETRS.DGETRS("N", n, 1, A, ipiv, rhs, info);
      if(info.val!=0)
         throw new RuntimeException("Badness in DGETRS LAPACK call, info="+info);
      double[] x=new double[n];
      for(int i=0; i<n; i++)
         x[i]=rhs[i][0]*Math.pow((c-1.0)/c, deaths[i]);
      return x;
   }
}
